import re
import sys
import os
import hashlib

def isValidProof(fprint, proof, difficulty):
    keyhash = hashlib.sha256()
    fingerproof = fprint + '_' + str(proof)
    keyhash.update(fingerproof.encode('utf-8'))
    hashResult = keyhash.hexdigest()
    if not hashResult.startswith(difficulty * '0'):
        return False
    return True

def isValidSignaturesDict(d):
    for key, sig in d.items():
        if type(key) is not str or type(sig) is not str:
            return False
        if not isValidFingerprint(key):
            return False
    return True
    
def isValidFingerprint(s):
    res = re.match('^[a-fA-F0-9]{16}$', s)
    if res == None:
        return False
    return True

def isValidMD5Sum(s):
    res = re.match('^[a-f0-9]{32}$', s)
    if res == None:
        return False
    return True

def isValidLedgerKey(s):
    fprint = s[:16]
    if not isValidFingerprint(fprint):
        return False
    try:
        if s[16] != '_':
            return False
    except IndexError:
        return False
    proof = s[17:]
    try:
        proofint = int(proof)
    except ValueError:
        return False
    if not isValidProof(fprint, proofint, 6):
        return False
    return True

# Checks for the ledger's format since it's of dict type
def isValidLedger(ledger):
    if type(ledger) is not dict:
        return False
    for key, value in ledger.items():
        if type(key) is not str:
            return False
        if type(value) is not int:
            return False
        if not isValidLedgerKey(key):
            return False
        # Check if balance is 0 or above
        if value < 0:
            return False
        # Check if the sum of all balances is 100000000 times the number of keys
        if len(ledger) != 0:
            numOfKeys = len(ledger)
            balancesSum = 0
            for balance in ledger.values():
                balancesSum += balance
            if balancesSum / numOfKeys != 10000000:
                return False
    return True


class PacketParser:
    def __init__(self, packet = None):
        self.packet = packet
        self.type = None
        self.data = None
        self.errorMessage = None
    
    def setPacket(self, packet):
        self.packet = packet

    def getData(self):
        return self.data

    def getType(self):
        return self.type

    def isPacketValid(self):
        # Check for dict type
        if type(self.packet) is not dict:
            self.errorMessage = 'It is not dict'
            return False
        # Check for packet keys
        if not set(['Type', 'Data']) == set(list(self.packet.keys())):
            self.errorMessage = 'Wrong keys in the packet'
            return False

        # Check for correct types
        validTypes = ['HELLO', 'REQ_LEDGER', 'RESP_LEDGER', 'REQ_INTRO_KEY',
                      'RESP_INTRO_KEY', 'REQ_INTRO_KEY_END', 'REQ_PAY',
                      'RESP_PAY', 'REQ_PAY_END']
        if self.packet['Type'] not in validTypes:
            self.errorMessage = 'Invalid type'
            return False
        
        self.type = self.packet['Type']
        self.data = self.packet['Data']
        # Check for each type exclusively
        # First check if Data is list because only type HELLO has Data of type list
        if type(self.data) is list:
            # HELLO
            if self.type == 'HELLO':
                for val in self.data:
                    if type(val) is not dict:
                        self.errorMessage = 'HELLO: element in Data list is not dict'
                        return False
                    if not set(['Fingerprint', 'ProofOfWork']) == set(list(val.keys())):
                        self.errorMessage = 'HELLO: Wrong Data keys'
                        return False
                    if type(val['Fingerprint']) is not str:
                        self.errorMessage = 'HELLO: Fingerprint value is not a string'
                        return False
                    if not isValidFingerprint(val['Fingerprint']):
                        self.errorMessage = 'HELLO: Fingerprint value has not valid format'
                        return False
                    if type(val['ProofOfWork']) is not int:
                        self.errorMessage = 'HELLO: ProofOfWork value is not int'
                        return False
                    if val['ProofOfWork'] < 0:
                        self.errorMessage = 'HELLO: ProofOfWork is negative'
                        return False
        
        elif type(self.data) is dict:     
            
            # REQ_LEDGER
            if self.type == 'REQ_LEDGER':
                if len(self.data) != 0:
                    self.errorMessage = 'REQ_LEDGER: Data is not empty'
                    return False

            # RESP_LEDGER
            if self.type == 'RESP_LEDGER':
                if not set(['Ledger', 'Signatures']) == set(list(self.data.keys())):
                    self.errorMessage = 'RESP_LEDGER: Wrong Data keys'
                    return False
                if not isValidLedger(self.data['Ledger']):
                    self.errorMessage = 'RESP_LEDGER: Invalid ledger'
                    return False
                if not isValidSignaturesDict(self.data['Signatures']):
                    self.errorMessage = 'RESP_LEDGER: fprints-signatures dict invalid'
                    return False
                
            # REQ_INTRO_KEY
            if self.type == 'REQ_INTRO_KEY':
                if not set(['Key', 'Checksum', 'Sig']) == set(list(self.data.keys())):
                    self.errorMessage = 'REQ_INTRO_KEY: Wrong Data keys'
                    return False
                if not isValidLedgerKey(self.data['Key']):
                    self.errorMessage = 'REQ_INTRO_KEY: Wrong ledger key'
                    return False
                if not isValidMD5Sum(self.data['Checksum']):
                    self.errorMessage = 'REQ_INTRO_KEY: Wrong md5sum format'
                    return False
                if type(self.data['Sig']) is not str:
                    self.errorMessage = 'REQ_INTRO_KEY: The signature is not string'
                    return False

            # RESP_INTRO_KEY and REQ_INTRO_KEY_END
            if self.type == 'RESP_INTRO_KEY' or self.type == 'REQ_INTRO_KEY_END':
                if not set(['Checksum', 'Signatures']) == set(list(self.data.keys())):
                    self.errorMessage = self.type + ': Wrong Data keys'
                    return False
                if not isValidMD5Sum(self.data['Checksum']):
                    self.errorMessage = self.type + ': Wrong md5sum format'
                    return False
                if not isValidSignaturesDict(self.data['Signatures']):
                    self.errorMessage = self.type + ': fprints-signatures dict invalid'
                    return False
                    
            # REQ_PAY
            if self.type == 'REQ_PAY':
                if not set(['Fromkey', 'Tokey', 'Amount', 'Checksum', 'Sig']) == set(list(self.data.keys())):
                    self.errorMessage = 'REQ_PAY: Wrong Data keys'
                    return False
                if not isValidFingerprint(self.data['Fromkey']): 
                    self.errorMessage = 'REQ_PAY: Invalid Fromkey value'
                    return False
                if not isValidFingerprint(self.data['Tokey']): 
                    self.errorMessage = 'REQ_PAY: Invalid Tokey value'
                    return False
                if not type(self.data['Amount']) is float:
                    self.errorMessage = 'REQ_PAY: Amount is not float'
                    return False
                if self.data['Amount'] <= 0:
                    self.errorMessage = 'REQ_PAY: Amount is not larger than 0'
                    return False
                if not isValidMD5Sum(self.data['Checksum']):
                    self.errorMessage = 'REQ_PAY: Invalid checksum value'
                    return False
                if not type(self.data['Sig'] is str):
                    self.errorMessage = 'REQ_PAY: Signature is not string'
                    return False
                    
            # RESP_PAY and REQ_PAY_END
            if self.type == 'RESP_PAY' or self.type == 'REQ_PAY_END':
                if not set(['Checksum', 'Signatures']) == set(list(self.data.keys())):
                    self.errorMessage = self.type + ': Wrong Data keys'
                    return False
                if not isValidMD5Sum(self.data['Checksum']): 
                    self.errorMessage = self.type + ': Invalid checksum value'
                    return False
                if not isValidSignaturesDict(self.data['Signatures']):
                    self.errorMessage = self.type + ': fprints-signatures dict invalid'
                    return False
        else:
            return False
        return True




import socket

def simpleSend(data, remoteaddrs, port, waitResponse=False, timeout=None):
    result = []
    for addr in remoteaddrs:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((addr, port))
            s.sendall(data.encode('utf-8'))
            if waitResponse == True:
                s.settimeout(timeout)
                response = str(s.recv(1024), 'utf-8')
                result.append(response)
        except (socket.timeout, OSError):
            continue
        finally:
            s.close()
    return result




if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('data', type=str, help='Text to send')
    parser.add_argument('remoteaddrs', type=str, help='Comma separated ips')
    parser.add_argument('port', type=int, help='Port to connect')
    parser.add_argument('--wait', dest='wait', action='store_true')
    parser.add_argument('--no-wait', dest='wait', action='store_false')
    parser.set_defaults(wait=True)
    parser.add_argument('timeout', type=float, default=None, help='Timeout if waiting for response')
    args = parser.parse_args()

    remotes = set(args.remoteaddrs.split(','))
    simpleSend(args.data, remotes, args.port, args.wait, args.timeout)




import gnupg
import json
import hashlib
import re

def getmd5(data):
    datahash = hashlib.md5()
    datahash.update(data.encode('utf-8'))
    return datahash.hexdigest()

# Arguments
# 1: Path of GPG folder to use
# 2: The keys stored in the private_keys.json file
# 3: The keys we want to sign with
# 4: The data to sign
# 5: The password
def signWithKeys(logging, gpgdir, privateKeys, keysToUse, data, password):
    if type(data) is dict:
        data = json.dumps(data, sort_keys=True)
    gpg = gnupg.GPG(gnupghome=gpgdir, use_agent=False)
    signaturesDict = {}
    existingKeys = []
    # Check for existing keys
    for searchingKey in keysToUse:
        if searchingKey in privateKeys:
            for listedKey in gpg.list_keys(True):
                if searchingKey == listedKey['keyid']:
                    existingKeys.append(searchingKey)
    
    for key in existingKeys:
        signedData = gpg.sign(data, keyid=key,
                    passphrase=password)
        signaturesDict[key] = str(signedData)
    return signaturesDict
    

# Returns tuple
# First element is a list with the valid keys
# Second element is a dict with valid keys:signatures
# Arguments:
# keySigs: a dictionary with keys as keys and signatures as values
def getKeysThatSignedData(logging, gpgdir, keySigs, data):
    logHeader = '@getKeysThatSignedData: '
    validKeys = []
    validKeysSigs = {}
    gpg = gnupg.GPG(gnupghome=gpgdir, use_agent=False)
    for fprint in keySigs:
        for key in gpg.list_keys():
            if key['keyid'] == fprint:
                foundKey = True
                break
        if not 'foundKey' in locals():
            continue
        signature = keySigs[fprint]
        verification = gpg.verify(signature)
        if verification.key_id != fprint:
            logging.info(logHeader + 'Wrong verification keyid')
            continue
        if verification.status != 'signature valid':
            logging.info(logHeader + 'Not valid signature')
            continue
        messagePattern = re.compile('(?<=\n\n)\w+')
        extractFromSignature = messagePattern.search(signature)
        if extractFromSignature is None:
            logging.info(logHeader + 'Wrong signature format')
            continue
        result =  extractFromSignature.group(0)
        if result != data:
            logging.info(logHeader + 'Signed data is not the expected')
            logging.info(logHeader + 'Whole signature: {}'.format(signature))
            logging.info(logHeader + 'Result: {}'.format(result))
            logging.info(logHeader + 'dataToCheck: {}'.format(data))
            continue
        validKeys.append(fprint)
        validKeysSigs[fprint] = signature
    return validKeys, validKeysSigs






import os
import json


def init(kwargs):
    # Create .minicash folder if not existing and enter it
    try:
        os.chdir(kwargs['Homedir'])
        if not os.path.isdir('.minicash'):
            os.mkdir('.minicash')
        os.chdir(kwargs['Homedir'] + '/.minicash')
    except (OSError, PermissionError)as e:
        return {'Fail':{'Reason': str(e)}}

    # Create configuration file
    if not os.path.isfile('config.json'):
        config = {
            'PEER_SERVER': {'Ip': 'saav29.ddns.net', 'Port': '9999'},
            'KEY_SERVERS': {'adresses': [
                'pgp.mit.edu',
                'sks-keyservers.net',
                'pool.sks-keyservers.net',
                'eu.pool.sks-keyservers.net'
            ]}
        }
        jsonedConfig = json.dumps(config, indent=4)
        try:
            with open('config.json', 'w') as conffile:
                conffile.write(jsonedConfig)
        except (OSError, PermissionError)as e:
            return {'Fail':{'Reason': str(e)}}

    # Take care of key files and .gnupg folder
    try:
        if not os.path.isfile('private_keys.json'):
            with open('private_keys.json', 'w') as infile:
                infile.write('{}')
        if not os.path.isfile('peers.json'):
            with open('peers.json', 'w') as infile:
                infile.write('{}')
        if not os.path.isfile('ledger.json'):
            with open('ledger.json', 'w') as infile:
                infile.write('{}')
        if not os.path.isfile('minicash.log'):
            with open('minicash.log', 'w') as infile:
                infile.write('')
        if not os.path.isdir('.gnupg'):
            os.mkdir('.gnupg')
            os.chmod('.gnupg', 0o700)
    except (OSError, PermissionError)as e:
        return {'Fail':{'Reason': str(e)}}

    return {'Success': {}}




import hashlib

def getmd5(data):
    datahash = hashlib.md5()
    datahash.update(data.encode('utf-8'))
    return datahash.hexdigest()

def getsha256(data):
    datahash = hashlib.sha256()
    datahash.update(data.encode('utf-8'))
    return datahash.hexdigest()

def isValidProof(fprint, proof, difficulty):
    hashResult = getsha256(fprint + '_' + str(proof))
    if not hashResult.startswith(difficulty * '0'):
        return False
    return True




import asyncio
import json

class RequestProtocol(asyncio.Protocol):
    def __init__(self, message):
        self.message = message

    def connection_made(self, transport):
        self.transport = transport
        messageJson = json.dumps(self.message)
        transport.write(messageJson.encode('utf-8'))
        transport.write_eof()

    
class RequestResponseProtocol(asyncio.Protocol):
    def __init__(self, future, message):
        self.future = future
        self.message = message
        self.fullMessage = bytes()
        self.response = bytes()

    def connection_made(self, transport):
        self.transport = transport
        messageJson = json.dumps(self.message)
        self.transport.write(messageJson.encode('utf-8'))
        self.transport.write_eof()

    def data_received(self, data):
        self.fullMessage += data

    def eof_received(self):
        try:
            res = json.loads(self.fullMessage.decode('utf-8'))
        except json.decoder.JSONDecodeError:
            res = None
        self.future.set_result(res)
    
    def connection_lost(self, exc):
        self.transport.close()

    
def sendReceiveToMany(message, ips):
    async def connect(ip, loop):
        future = asyncio.Future()
        try:
            await loop.create_connection(lambda: RequestResponseProtocol(future, message), ip , 2222)
        except (OSError, ConnectionRefusedError):
            return
        await future
        return future

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    tasks = []
    for ip in ips:
        task = asyncio.ensure_future(connect(ip, loop))
        tasks.append(task)
    results = loop.run_until_complete(asyncio.gather(*tasks))   
    loop.close()
    while None in results:
        results.remove(None)
    rawResults = []
    for res in results:
        rawResults.append(res.result())
    return rawResults


def sendToMany(message, ips):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    tasks = []
    for ip in ips:
        task = loop.create_connection(lambda: RequestProtocol(message), ip , 2222) 
        tasks.append(task)
    try:
        results = loop.run_until_complete(asyncio.gather(*tasks))   
    except ConnectionRefusedError:
        pass
    loop.close()




from multiprocessing import Pool, cpu_count
import hashlib
import signal

def checkHash(argsList):
    key = argsList[0]
    pattern = argsList[1]
    cpusNum = argsList[2]
    coreId = argsList[3]
    while True:
        coreId += cpusNum
        newHash = hashlib.sha256()
        newFeed = key + '_' + str(coreId)
        newHash.update(newFeed.encode('utf-8'))
        result = newHash.hexdigest()
        if result.startswith(pattern):
            return coreId

def signalIgnorer():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


class POWGenerator:
    def __init__(self, key, difficulty, requestedCores):
        self.key = key
        self.beginningPattern = '0' * difficulty
        coresNum = cpu_count()
        if requestedCores <= coresNum:
            self.coresToUse = requestedCores
        else:
            self.coresToUse = coresNum
    
    def getSolution(self):
        jobsArgsList = []
        for cou in range(0,self.coresToUse):
            jobsArgsList.append((self.key, self.beginningPattern, self.coresToUse, cou))

        with Pool(self.coresToUse, signalIgnorer) as workPool:
            try:
                for gotmatch in workPool.imap_unordered(checkHash, jobsArgsList):
                    if gotmatch:
                        return gotmatch
            except KeyboardInterrupt:
                # exit()
                return None

if __name__ == "__main__":
    import sys
    key = sys.argv[1]
    difficulty = sys.argv[2]
    cores = sys.argv[3]
    gen = POWGenerator(key, int(difficulty), int(cores))
    print("Solution: {}".format(gen.getSolution()))



from setuptools import setup

setup(name='Minicash',
    version='0.4.6',
    description='A cryptocurrency without blockchain',
    url='http://github.com/saavedra29/Minicash',
    author='Aristides Tomaras',
    author_email='arisgold29@gmail.com',
    license='MIT',
    scripts=['minicash/minicash', 'minicash/minicashd', 'minicash/quickDataGen'], 
    packages=['minicash', 'minicash.utils'],
    install_requires=[
      'json-rpc','python-daemon', 'python-gnupg', 'lockfile'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Security :: Cryptography'
    ],
    keywords='Cryptocurrency gpg',
    zip_safe=False)
